/** An implementation of Rust's `Result` type in TypeScript. */
export declare class Result<T, E> {
    private value;
    constructor(value: ["ok", T] | ["err", E]);
    isOk(): boolean;
    isErr(): boolean;
    unwrap(): T;
    unwrapErr(): E;
    expect(message: string): T;
    expectErr(message: string): E;
    map<U>(f: (value: T) => U): Result<U, E>;
    mapErr<F>(f: (value: E) => F): Result<T, F>;
    /** Fixes the typings on the `Err` variant of a Result. Unsafe operation.
     *
     *  # Example
     *  ```
     *  function a(): Result<string, string> {
     *      return ok("good news!");
     *  }
     *
     *  function b(): Result<string, number> {
     *      const result = a();
     *      if (result.isErr()) {
     *          return err(10);
     *      } else {
     *          // Without fix: Type error: string is not assignable to number
     *          return result.fixErr();
     *      }
     *  }
     *  ```
     */
    fixErr<B>(): Result<T, B>;
    /** Fixes the typings on the `Ok` variant of a Result. Throws if this value
     *  is an `Ok`.
     *
     *  # Example
     *  ```
     *  function a(): Result<string, string> {
     *      return err("oh no!");
     *  }
     *
     *  function b(): Result<number, string> {
     *      const result = a();
     *      if (result.isErr()) {
     *          // Without fix: Type error: string is not assignable to number
     *          return result.fixOk();
     *      } else {
     *          return ok(5);
     *      }
     *  }
     *  ```
     */
    fixOk<A>(): Result<A, E>;
    /** Similar to Rust's `match`. */
    match<U>(matcher: {
        ok: (value: T) => U;
        err: (error: E) => U;
    }): U;
    /** Flattens a `Result<Result<T, E>, E>` into a `Result<T, E>`. */
    flatten: T extends Result<infer U, E> ? () => Result<U, E> : never;
    /** Returns `undefined` on error, else the inner value. */
    undefined(): T | undefined;
    /** Throws `E` or returns `T`. */
    throw(): T;
}
export declare function ok<const T, E = never>(value: T): Result<T, E>;
export declare function ok<T, E = never>(value: T): Result<T, E>;
export declare function err<T = never, const E = unknown>(value: E): Result<T, E>;
export declare function err<T = never, E = unknown>(value: E): Result<T, E>;
