/** An implementation of Rust's `Result` type in TypeScript. */
export class Result {
    value;
    constructor(value) {
        this.value = value;
    }
    isOk() {
        return this.value[0] === "ok";
    }
    isErr() {
        return this.value[0] === "err";
    }
    unwrap() {
        if (this.isOk()) {
            return this.value[1];
        }
        else {
            throw new Error(`Unwrapping error result: ${JSON.stringify(this.value[1])}`);
        }
    }
    unwrapErr() {
        if (this.isErr()) {
            return this.value[1];
        }
        else {
            throw new Error(`Unwrapping successful result: ${JSON.stringify(this.value[1])}`);
        }
    }
    expect(message) {
        if (this.isOk()) {
            return this.value[1];
        }
        else {
            throw new Error(message + `: ${JSON.stringify(this.value[1])}`);
        }
    }
    expectErr(message) {
        if (this.isErr()) {
            return this.value[1];
        }
        else {
            throw new Error(message);
        }
    }
    map(f) {
        if (this.isOk()) {
            return ok(f(this.unwrap()));
        }
        else {
            return err(this.unwrapErr());
        }
    }
    mapErr(f) {
        if (this.isErr()) {
            return err(f(this.unwrapErr()));
        }
        else {
            return ok(this.unwrap());
        }
    }
    /** Fixes the typings on the `Err` variant of a Result. Unsafe operation.
     *
     *  # Example
     *  ```
     *  function a(): Result<string, string> {
     *      return ok("good news!");
     *  }
     *
     *  function b(): Result<string, number> {
     *      const result = a();
     *      if (result.isErr()) {
     *          return err(10);
     *      } else {
     *          // Without fix: Type error: string is not assignable to number
     *          return result.fixErr();
     *      }
     *  }
     *  ```
     */
    fixErr() {
        const value = this.expect("cannot use `fixErr` on `Err` Result");
        return ok(value);
    }
    /** Fixes the typings on the `Ok` variant of a Result. Throws if this value
     *  is an `Ok`.
     *
     *  # Example
     *  ```
     *  function a(): Result<string, string> {
     *      return err("oh no!");
     *  }
     *
     *  function b(): Result<number, string> {
     *      const result = a();
     *      if (result.isErr()) {
     *          // Without fix: Type error: string is not assignable to number
     *          return result.fixOk();
     *      } else {
     *          return ok(5);
     *      }
     *  }
     *  ```
     */
    fixOk() {
        const error = this.expectErr("cannot use `fixOk` on `Ok` Result");
        return err(error);
    }
    /** Similar to Rust's `match`. */
    match(matcher) {
        if (this.isOk()) {
            return matcher.ok(this.unwrap());
        }
        else {
            return matcher.err(this.unwrapErr());
        }
    }
    /** Flattens a `Result<Result<T, E>, E>` into a `Result<T, E>`. */
    flatten = (() => {
        // eslint-disable-next-line @typescript-eslint/no-unsafe-return
        return this.match({
            err: (e) => err(e),
            ok: (v) => {
                if (v instanceof Result) {
                    // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-return
                    return v;
                }
                else {
                    return ok(v);
                }
            }
        });
    });
    /** Returns `undefined` on error, else the inner value. */
    undefined() {
        return this.match({
            ok: (v) => v,
            err: () => undefined
        });
    }
    /** Throws `E` or returns `T`. */
    throw() {
        return this.match({
            ok: (v) => v,
            err: (e) => {
                throw e;
            }
        });
    }
}
export function ok(value) {
    return new Result(["ok", value]);
}
export function err(value) {
    return new Result(["err", value]);
}
//# sourceMappingURL=result.js.map