type KeyResult = {
    /** How important this key is relative to other keys. Higher = closer to top.
     *
     *  For example, one might want to prioritize the name of a thing over its description.
     *  Suppose the name has a bias of 3 and the description has a bias of 1,
     *  then even if the description was an exact match, if the name starts with
     *  the query, the name would be prioritized over the description. However,
     *  if it was a 'character appears in order' match, the description would be prioritized.
     */
    bias: number;
    /** The actual string to compare against. */
    value: string;
    /** The minimum (inclusive, unweighted) priority. If something is less than
     *  this, it is discarded.
     *
     *  Preferably use something from `SEARCH_PRIORITIES`.
     *
     *  @default 0
     */
    minPriority?: number;
};
export declare function filterByString<T>(searchQuery: string, data: Array<T> | undefined, key: (item: T) => string | Array<KeyResult>, filter?: (item: T) => boolean): Array<T>;
export declare const SEARCH_PRIORITIES: {
    /** The search query is an exact match to the string.
     *  E.g., `pi` matches '**Pi**'
     */
    readonly EXACT_MATCH: 6;
    /** The search query appears, in full, at the start of the string as the
     *  *full* word.
     *  E.g., `pi` matches '**Pi** Shop'
     */
    readonly STARTS_WITH_FULL_WORD: 5;
    /** The search query appears, in full, at the start of the string.
     *  E.g., `pi` matches '**Pi**zza Card'
     */
    readonly STARTS_WITH: 4;
    /** The search query appears, in full, after a space (thus denoting the
     *  beggining of a word.)
     *  E.g., `pi` matches 'Papa's **Pi**zzeria'
     */
    readonly STARTS_WITH_AFTER_SPACE: 3;
    /** The search query appears, in full, somewhere in the string.
     *  E.g., `pi` matches 'Pop**pi**ng In'
     */
    readonly INCLUDES: 2;
    /** The characters in the search query appear, in order, in the string.
     *  E.g., `pi` matches '**P**ay **i**t Now'
     */
    readonly CHARACTERS_IN_ORDER: 1;
    /** The search query does not appear in the string.
     *  E.g., `pi` does not match 'Glopsop's Mop Shop'
     */
    readonly NO_MATCH: 0;
};
export {};
